# Copyright (C) 2001  Miklos Reiter (see the file README for details on
# how to contact me)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

# deepdep.awk
#
# This awk program calculates the deep (indirect) dependencies of java files, from the direct
# dependencies, as determined by depend.awk.
#
# Inputs:
# - classes: a file, where each $1 is the fully qualified name of a java class, $2 is the name of the 
#   java file and $3 is the name of the .dep file (in ./src); this is the "classlist" format as produced by
#   javafile2classlistand; this file contains all java classes relevant for dependency checking
#   (i.e. all classes in the project)
# - changed: classlist format; this list represents those classes whose source files have changed, and 
#   which thus need to be recompiled.
# All further arguments are treated as a further source of changed classes, just as changed.
#
#
# Output:
# - a list of classes to be recompiled, this includes all classes in changed and those that, directly or
#   indirectly, depend on them
#



#--------------------------------------------INIT-----------------------------------------------


# Store the name of the classes file, to be able to use special rule for this file.
BEGIN {
  classes_file = ARGV[1];
}

# Rule to process first argument specially, calls next, so no further rules get processed.
(FILENAME == classes_file) {
  class_name = $1;
  classes[class_name] = FNR; # index of class, indexed by (fully qualified) name
  classlistentry[FNR] = $0; # complete info, as needed for output
  depfiles[FNR] = $3; # needed in END rule, when processing is done
  num_classes++;
  next
}


# Rule to process for other files (changed).
{
  if ($1 in classes)
  {
    class_name = $1;
    i = classes[class_name];
    newly_reached[i] = 1
  }
  else print ("Unknown class " $1 ", ignoring.") > "/dev/stderr"
}



#--------------------------------------------PROCESSING-------------------------------------------

# reads direct dependencies from the dependency files of all known classes, and stores them in
# deptable, a two-dimensional array indexed by integers from 1
function read_dependencies()
{
  for (class in classes)
  {
    i = classes[class];
    # open depfiles[i] and read dependent classes, don't complain about non-existing files
    while ((getline < (depfiles[i])) > 0)
    {
      class_name = $1;
      class_index = classes[class_name];
      deptable[class_index, i] = 1
    }
    close(depfiles[i])
  }
}


# returns 1 if newly_reached contains at least one element, 0 if it's empty
function non_empty_newly_reached()
{
  for (i in newly_reached) return 1;
  return 0
}


# This calculates the indirect dependencies (classes "reached by the changed ones").
# The algorithm used is O(n^2), thus more efficient than Warshall's algorithm, which is
# O(n^3). This is possible, because our problem is more specific than finding the
# transitive closure of the dependency relation. We merely want to find those classes,
# which are related to one of the changed classes by the transitive closure of the
# dependency relation.
function get_reached()
{
  # We use the following arrays:
  # ((i, j) in deptable) iff classes[i] depends on classes[j] (set up by read_dependencies)
  # (i in reached) iff we know that classes[i] depends on one of the changed classes
  # those whose index was initially set in newly_reached
  # (i in newly_reached) iff classes[i] has been "reached" in the last iteration
  while (non_empty_newly_reached())
  {
    # reached = (reached OR newly_reached)
    for (i in newly_reached)
    {
      reached[i] = 1
    }

    # newly_reached = direct_dependents(newly_reached)
    split("", old_newly_reached); # portable way of deleting an array
    for (i in newly_reached)
    {
      old_newly_reached[i] = 1;
    }
    split("", newly_reached);
    for (i in old_newly_reached)
    {
      for (j = 1; j <= num_classes; j++)
      {
        if (deptable[i, j]) newly_reached[j] = 1;
      }
    }

    # newly_reached = newly_reached AND NOT reached
    for (i in reached)
    {
      delete newly_reached[i];
    }
  }
}


# output the complete class info (classlist format) of classes whose index is in reached
function output_reached()
{
  for (i in reached)
  {
    print classlistentry[i]
  }
}


# Rule doing all the processing
END {
  read_dependencies(); # read the direct dependencies from the .dep files
  get_reached(); # calculate indirect dependencies (classes "reached by the changed ones")
  output_reached();
}
