/**
 * Copyright (C) 2001  Miklos Reiter (see the file README for details on
 * how to contact me)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package gravsphcolldemo;

import noope.input.Valid;
import noope.input.ArrayBlockReader;

import noope.core.Vector3D;

import java.lang.IllegalArgumentException;

import java.awt.Color;

/**
 *
 * This class stores information needed to construct an ArrayBlockReader defining an Entity for
 * a simulation in Applet2. It is used to transfer information between Applet2 and
 * EditEntityDialog.
 *
 * @author Miklos Reiter
 * @version 27-Apr-2001
 */
public class EntityBrData
{
    private String name, mass, radius, position, velocity, color;

    /**
     * Constructs a new EntityBrData object without initializing any of its data members.
     */
    public EntityBrData()
    {
    }

    /**
     * Constructs a new EntityBrData object from the given data by calling the setXXXX methods.
     */
    public EntityBrData(String aName, String aMass, String aRadius, String aPosition,
			String aVelocity, String aColor) throws IllegalArgumentException
    {
	setName(aName);
	setMass(aMass);
	setRadius(aRadius);
	setPosition(aPosition);
	setVelocity(aVelocity);
	setColor(aColor);
    }

    /** Returns the name. */
    public String getName()
    {
	return name;
    }

    /** Sets the name. */
    public void setName(String aName)
    {
	name = aName;
    }

    /** Returns the mass. */
    public String getMass()
    {
	return mass;
    }

    /** Sets the mass. */
    public void setMass(String aMass) throws IllegalArgumentException
    {
	if (Valid.validDouble(aMass))
		    mass = aMass;
	else throw new IllegalArgumentException();
    }

    /** Returns the radius. */
    public String getRadius()
    {
	return radius;
    }

    /** Sets the radius. */
    public void setRadius(String aRadius) throws IllegalArgumentException
    {
	if (Valid.validDouble(aRadius))
		    radius = aRadius;
	else throw new IllegalArgumentException();
    }

    /** Returns the position. */
    public String getPosition()
    {
	return position;
    }

    /** Sets the position. */
    public void setPosition(String aPosition) throws IllegalArgumentException
    {
	if (Valid.validVector3D(aPosition))
		    position = aPosition;
	else throw new IllegalArgumentException();
    }

    /** Gets the velocity. */
    public String getVelocity()
    {
	return velocity;
    }

    /** Sets the velocity. */
    public void setVelocity(String aVelocity) throws IllegalArgumentException
    {
	if (Valid.validVector3D(aVelocity))
		    velocity = aVelocity;
	else throw new IllegalArgumentException();
    }

    /** Gets the Color. */
    public String getColor()
    {
	return color;
    }

    /** Sets the Color. */
    public void setColor(String aColor) throws IllegalArgumentException
    {
	if (Valid.validColor(aColor))
		    color = aColor;
	else throw new IllegalArgumentException();
    }

    /**
     * Asserts that the data stored is complete (i.e. everything has been set).
     */
    public void assertValid() throws IllegalArgumentException
    {
	if (name == null || mass == null || radius == null || position == null || velocity == null
	    || color == null)
	    throw new IllegalArgumentException();
    }

    /**
     * Returns a String representation of the data that can be used, for example, to store this
     * object in an AWT List.
     */
    public String toString() throws IllegalArgumentException
    {
	assertValid();
	return ("Name: " + name + "; m = " + mass + "; r = " + radius + "; x = " + position
		+ "; v = " + velocity + "; color = " + color);
    }

    /** creates an ArrayBlockReader defining an Entity */
    private ArrayBlockReader newEntityBr(String[] projectionLawNames, int numberOfOutputProjectionLaws, 
				    String name, String position, String velocity,
				    String inertialMass, String radius, String colorRGB)
    {
	String header = "entity", parameter = "";
	String[][] entries = 
	{{"name", name},
	 {"position", position},
	 {"velocity", velocity},
	 {"inertialMass", inertialMass}
	};
	ArrayBlockReader[] subblocks = new ArrayBlockReader[1+numberOfOutputProjectionLaws];
	subblocks[0] = newSphereCollPropBr(radius);
	
	for (int i = 1; i <= numberOfOutputProjectionLaws; i++)
	    {
		subblocks[i] = 
		    new ArrayBlockReader(projectionLawNames[i-1], "OutputProjectionCirclePropertyRecord",
					 new String[][]{{"colorRGB", colorRGB},
							{"radius", radius}
					 }, null);
	    }
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }
    
    /** creates an ArrayBlockReader defining a SphereCollisionPropertyRecord */
    private ArrayBlockReader newSphereCollPropBr(String radius) 
    {
	String header = "SphereCollision", parameter = null; 
	String[][] entries = {
	    {"radius", radius}
	};
	ArrayBlockReader[] subblocks = {};

	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }

  private static Color string2Color(String s)
  {
    if (s.equals("black")) return Color.black; else if
           (s.equals("blue")) return Color.blue; else if
           (s.equals("cyan")) return Color.cyan; else if
           (s.equals("darkGray")) return Color.darkGray; else if
           (s.equals("gray")) return Color.gray; else if
           (s.equals("green")) return Color.green; else if
           (s.equals("lightGray")) return Color.lightGray; else if
           (s.equals("magenta")) return Color.magenta; else if
           (s.equals("orange")) return Color.orange; else if
           (s.equals("pink")) return Color.pink; else if
           (s.equals("red")) return Color.red; else if
           (s.equals("white")) return Color.white; else if
           (s.equals("yellow")) return Color.yellow;
    else
      try
      {
        Vector3D v = new Vector3D(s);
        return new Color((float)v.getX(), (float)v.getY(), (float)v.getZ());
      }
      catch (java.text.ParseException e)
      {
	  return null;
      }
  }


    /**
     * First calls assertValid() and then returns an ArrayBlockReader representing the Entity
     * this EntityBrData object represents.
     *
     * @arg projLawNames The array storing the names of the OutputProjection laws used (for which
     * we will need to create LawPropertyRecords).
     * @arg projLawNum The number of OutputProjection laws present.
     */
    public ArrayBlockReader getArrayBlockReader(String[] projLawNames, int projLawNum) 
	throws IllegalArgumentException
    {
	assertValid();
	String sColor; // string storing a Color in RGB form (with components in range 0.0-1.0)
	Color col; // stores a Color
	
	assertValid();

	col = string2Color(getColor());
	if (col == null) throw new IllegalArgumentException("Invalid Color");
	sColor = ("(" + ((float)col.getRed()/255.0F) + "," + ((float)col.getGreen()/255.0F) +
		  "," + ((float)col.getBlue()/255.0F) + ")");
	
	
	return newEntityBr(projLawNames, projLawNum, getName(), getPosition(), getVelocity(),
			   getMass(), getRadius(), sColor);
    }
}
