/**
 * Copyright (C) 2001  Miklos Reiter (see the file README for details on
 * how to contact me)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package gravsphcolldemo;

import java.lang.IllegalArgumentException;

import noope.core.Vector3D;

import noope.input.Valid;
import noope.input.ArrayBlockReader;

/**
 *
 * This class stores data that can be used to construct an ArrayBlockReader
 * representing an OutputProjection law. It is used by Applet2 to transfer data
 * to and from EditProjectionDialog.
 *
 * @author Miklos Reiter
 * @version May 2001
 */

public class ProjectionBrData {

    private String name, normal, xmin, xmax, ymin, ymax;
    
    private String period = "0.01";
    
    /** Creates a new ProjectionBrData object, without initializing any of the variables. */
    public ProjectionBrData()
    {
    }
    
    /** Creates a new ProjectionBrData object that is initialized from the arguments.
     *
     * @arg aName The name of the OutputProjection law.
     * @arg aNormal The normal to the plane.
     * @arg aXmin The lowest (new) x-coordinate displayed.
     * @arg aXmax The highest (new) x-coordinate displayed.
     * @arg aYmin The lowest (new) y-coordinate displayed.
     * @arg aYmax The highest (new) y-coordinate displayed.
     */
    public ProjectionBrData(String aName, String aNormal, String aXmin, String aXmax, String aYmin,
			    String aYmax) throws IllegalArgumentException
    {
	setName(aName);
	setNormal(aNormal);
	setXmin(aXmin);
	setXmax(aXmax);
	setYmin(aYmin);
	setYmax(aYmax);
    }

  /**
   * Returns the name of the OutputProjection law.
   */
  public String getName()
  {
    return name;
  }

  /**
   * Sets the name of the OutputProjection law.
   */
  public void setName(String newName)
  {
    name = newName;
  }

  /**
   * Returns the normal.
   */
  public String getNormal()
  {
    return normal;
  }

  /**
   * Sets the normal.
   */
  public void setNormal(String newNormal) throws IllegalArgumentException
  {
      if (Valid.validVector3D(newNormal))
	  {
	      normal = newNormal;
	  }
      else throw new IllegalArgumentException();
  }

  /**
   * Returns xmin.
   */
  public String getXmin()
  {
    return xmin;
  }

  /**
   * Sets xmin.
   */
  public void setXmin(String newXmin) throws IllegalArgumentException
  {
      if (Valid.validDouble(newXmin))
	  {
	      xmin = newXmin;
	  }
      else throw new IllegalArgumentException();
  }

  /**
   * Returns xmax.
   */
  public String getXmax()
  {
    return xmax;
  }

  /**
   * Sets xmax.
   */
  public void setXmax(String newXmax) throws IllegalArgumentException
  {
      if (Valid.validDouble(newXmax))
	  {
	      xmax = newXmax;
	  }
      else throw new IllegalArgumentException();
  }

  /**
   * Gets ymin.
   */
  public String getYmin()
  {
    return ymin;
  }

  /**
   * Sets ymin.
   */
  public void setYmin(String newYmin) throws IllegalArgumentException
  {
      if (Valid.validDouble(newYmin))
	  {
	      ymin = newYmin;
	  }
      else throw new IllegalArgumentException();
  }

  /**
   * Gets ymax.
   */
  public String getYmax()
  {
    return ymax;
  }

  /**
   * Sets ymax
   */
  public void setYmax(String newYmax) throws IllegalArgumentException
  {
      if (Valid.validDouble(newYmax))
	  {
	      ymax = newYmax;
	  }
      else throw new IllegalArgumentException();
  }



    /**
     * Asserts that all the data is provided (If it is, it has to be valid, because there is no
     * way of storing it in this object, other than getting it via the public accessors.
     */
    public void assertValid() throws IllegalArgumentException
    {
	if ((name == null) || (normal == null) || (xmin == null) || (xmax == null)
	    || (ymin == null) || (ymax == null))
	    throw new IllegalArgumentException();
    }

    /**
     * This method returns a String representation of ProjectionBrData, that can be used to add it
     * to a AWT List.
     */
    public String toString() throws IllegalArgumentException
    {
	assertValid();
	return
	    ("Name: " + getName() + "; n = " + normal + "; x: " + xmin + " - " + 
	     xmax + "; y: " + ymin + " - " + ymax);
    }

    /** creates an ArrayBlockReader defining a OutputProjection law */
    private ArrayBlockReader newOutputProjectionBr(String name, String period, String normal, 
						     String xmin,
						     String xmax, String ymin, String ymax,
						     String width, String height, String xpos,
						     String ypos)
    {
	String header = "law", parameter = "noope.output.OutputProjection";
	String[][] entries =
	{{"name", name},
	 {"period", period},
	 {"normal", normal},
	 {"xmin", xmin},
	 {"xmax", xmax},
	 {"ymin", ymin},
	 {"ymax", ymax},
	 {"width", width},
	 {"height", height},
	 {"xpos", xpos},
	 {"ypos", ypos}
	};
	ArrayBlockReader[] subblocks = {};
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }

    /** creates an ArrayBlockReader defining a OutputProjection law */
    private ArrayBlockReader newOutputProjectionBr(String name, String period, String normal, 
						   String xmin,
						   String xmax, String ymin, String ymax,
						   String width, String height, String gifName)
    {
	String header = "law", parameter = "noope.output.OutputProjection";
	String[][] entries =
	{{"name", name},
	 {"period", period},
	 {"normal", normal},
	 {"xmin", xmin},
	 {"xmax", xmax},
	 {"ymin", ymin},
	 {"ymax", ymax},
	 {"width", width},
	 {"height", height},
	 {"gifName", gifName}
	};
	ArrayBlockReader[] subblocks = {};
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }
        
    /**
     * Returns an ArrayBlockReader object representing the OutputProjection law, whose data is
     * stored in this ProjectionBrData object. Calls assertValid() first.
     *
     * @arg x The x-coordinate of the location of the window.
     * @arg y The y-coordinate of the location of the window.
     */
    public ArrayBlockReader getArrayBlockReader(String x, String y)
	throws IllegalArgumentException
    {
	assertValid();
	return newOutputProjectionBr(getName(), period, getNormal(), getXmin(), getXmax(),
				     getYmin(), getYmax(), "300", "300", x, y);
    }

    /**
     * Returns an ArrayBlockReader object representing the OutputProjection law, whose data is
     * stored in this ProjectionBrData object. Calls assertValid() first.
     *
     * @arg x The x-coordinate of the location of the window.
     * @arg y The y-coordinate of the location of the window.
     * @arg period The sampling period of the OutputProjection law.
     */
    public ArrayBlockReader getArrayBlockReader(String x, String y, String period)
	throws IllegalArgumentException
    {
	assertValid();
	return newOutputProjectionBr(getName(), period, getNormal(), getXmin(), getXmax(),
				     getYmin(), getYmax(), "300", "300", x, y);
    }

    /**
     * Returns an ArrayBlockReader object representing the OutputProjection law, whose data is
     * stored in this ProjectionBrData object. Calls assertValid() first.
     *
     * @arg gifName The name of the gif file to write the animation to.
     * @arg period The sampling period of the OutputProjection law.
     */
    public ArrayBlockReader getAnimGifArrayBlockReader(String gifName, String period)
	throws IllegalArgumentException
    {
	assertValid();
	return newOutputProjectionBr(getName(), period, getNormal(), getXmin(), getXmax(),
				     getYmin(), getYmax(), "300", "300", gifName);
    }
}
