/**
 * Copyright (C) 2001  James Kermode (see the file README for details on
 * how to contact me)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


package noope.core;

import java.lang.Integer;
//import java.util.ListIterator;
import java.util.Vector;
import java.lang.reflect.Constructor;

import noope.laws.Law;
import noope.input.BlockReader;
import noope.input.BlockReaderLocation;
import noope.input.BRLoadingException;
import noope.input.BRLoadingParseException;
import noope.input.BREntryNotFoundException;
import noope.input.BRClassLoaderUtils;
import noope.input.LawLoader;

/** 
 * This class maintains a list of Law objects.
 *
 *  It is a subclass of Vector, and
 *  as such it implements the List interface. In addition to the normal Vector
 *  element accessor methods, Law objects are referable by a String name, using the
 *  getLaw(String) method - this is only used when initialising the
 *  LawPropertyRecords of Entity objects. The Law objects to be stored are
 *  constructed and added to the Laws object by the Physics() constructor, using a
 *  BlockReader.
 *
 * @author James Kermode
 * @version 28-Apr-2001 Removed ListIterator for compatibility with Java 1.1 [Miklos Reiter]
 */
public class Laws extends Vector
{

   /**
    * Constructs the Laws object from the supplied BlockReader.
    *
    * This should be the laws{...} block from the main BlockReader. This
    * constructor will only be called by the Physics(BlockReader) constructor.
    *
    * @param br BlockReader used to read in the Law objects
    * @param phys Physics object passed on to Law() constructors
    * @see noope.laws.Law#Law(BlockReader, int, Physics)
    * @exception BRLoadingException if invalid data is read from the BlockReader
    */
    public Laws(BlockReader br, Physics phys) throws BRLoadingException
    {
	LawLoader lawLoader = phys.getLawLoader();

	// Check it's not empty to start with
	if (!br.hasMoreBlocks())
	    throw new BRLoadingParseException("Can't construct Laws from empty BlockReader", 
	                                      br.getEntryLocation());

	BlockReaderLocation brL = null; // for BR*exceptions.

  	// Load laws one at a time from the BlockReader
	BlockReader brLaw;
	String lawClassName; // Class file to construct
	/* The current lawNumber: passed to the Law constructor. This has to be
	 * an Integer rather than an int so it can be put in the Object[] passed
	 * to the Law constructor
	 */

     
	int lawNumber = 0;

	while (br.hasMoreBlocks())
	{
	    brLaw = br.getNextBlock();
	    brL = br.getBlockLocation();

	    if (!brLaw.getHeader().equals("law")) // Check it's a law block
	       throw new BRLoadingParseException("Non-law block found in laws{...}", 
	                                          brLaw.getEntryLocation());
	    lawClassName = brLaw.getParameter();


	    this.addElement(lawLoader.createLaw(lawClassName, brL, brLaw, lawNumber, phys));

	    brLaw.assertEmpty();

	    lawNumber++;
	}
	br.assertEmpty();
    }

   /**
    * Return a Law identified by the a string, or throw a LawNotFoundException
    * if this Law does not exist.
    * @param lawname search for a Law whose getName() returns <i>lawname</i>
    * @exception LawNotFoundException if the Law is not found
    */
    public Law getLaw(String lawname) throws LawNotFoundException
    {
	Law law;
	// Otherwise, try to find a law with that name

	for (int i = 0; i < size(); i++)
	    {
		law = (Law)elementAt(i);
		if (law.getName().equals(lawname))
		    return law;
	    }

	// Not found, so throw a LawNotFoundException
	throw new LawNotFoundException();
    }

    /** Disposes of resources by calling dispose() on each of the Laws stored in the List. */
    public void dispose()
    {
	for (int i = 0; i < size(); i++)
	    {
		((Law)elementAt(i)).dispose();
	    }
	
	// remove all elements now
	this.removeAllElements();
    }
}
