/**
 * Copyright (C) 2001  Miklos Reiter (see the file README for details on
 * how to contact me)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package noope.entities;

import java.text.ParseException;
import java.lang.NumberFormatException; // thrown by Double.parseDouble
import java.lang.Double; // parseDouble
import java.lang.reflect.Array;

import noope.core.*;
import noope.laws.Law;
import noope.laws.LawPropertyRecord;
import noope.input.BlockReader;
import noope.input.BRLoadingException;
import noope.input.BRLoadingParseException;
import noope.input.BREntryNotFoundException;


/** This class represents a physical object modelled. Physics, the physics
  engine class, contains an object of the Entities class, which stores
  internally, a list of Entities.
  
  It also stores an internal forceAccumulator that is used by the Law 
  objects to accumulate their contibution to the total force on that Entity.

  @author Miklos Reiter 
  @version 19 April 2001: added dispose() [Miklos Reiter]
  */
  
public class Entity
{
    private MutableVector3D position, velocity, forceAccumulator;
    // forceAccumulator is used by the Law objects to accumulate their
    // contibution to the total force on that Entity.
    private double inertialMass;
    /** The name associated with this Entity */
    protected String name;
    /** An array of LawPropertyRecords that store the properties of this
      entity for each Law currently in use */
    protected LawPropertyRecord[] properties;
    /** A reference to the physics world so getActiveForce can learn about its
    environment. */
    protected Physics physics;
    
    /**
    Construcst an Entity from the supplied BlockReader.
    This constructor expects the BlockReader to contain the name as the first Entry; after that,
    position, velocity, inertialMass (the first two vectors of the form (x, y, z)
    and the latter just a number), in this order. If one is missing, the constructor
    will assume that it is not provided and will use the default value. If it
    then occurrs later, an exception is raised.
    @param br BlockReader used to initialise name, position, velocity, inertialMass and properties.
    @param phys Reference to the Physics world
    */
    public Entity(BlockReader br, Physics phys) throws BRLoadingException
    {
	physics = phys;
	forceAccumulator = new MutableVector3D(); // 0
	initDataMembers(br); // initialize data members from br
	initproperties(br); // initialize the properties (array of
	// LawPropertyRecord's.
    }
    
    /** Return this Entity's inertial mass, used in Newton's second law (F = ma) calculations */
    public double getInertialMass() { return inertialMass; }
    /** Return this Entity's position */
    public Vector3D getPosition() { return position; }
    /** Return this Entity's velocity */
    public Vector3D getVelocity() { return velocity; }
    
    /** Return this Entity's name */
    public String getName() { return name; }     
    
    /** This method gets the LawPropertyRecord corresponding to the law with the given number.
      as it's refernce number. Lawnumbers are allocated to each Law by the Laws object
      when they are created.
      @see noope.core.Laws#Laws(BlockReader, Physics)
      @see noope.laws.Law#constructNewPropertyRecord(BlockReader, Entity) */
    public LawPropertyRecord getPropertyRecord(int lawnumber)
    {
	return properties[lawnumber];
    }
    
    /** Apply a force to this Entity. The force is represented by the Vector3D <i>force</i>.
    This method is called by each law. It adds <i>force</i> to <i>forceAccumulator</i>. */
    public void addForce(Vector3D force)
    { // adds force to forceAccumulator
	forceAccumulator.addInplace(force);
    }
    
    /** Apply Newtonian physics to update velocity and position. Also resets accumulator and
    calls getActiveForce before it runs so it can use the active force too.
    No need to override. 
    @param deltaTime the time step to use when updating velocity and position */
    public void step(double deltaTime)
    {
	addForce(getActiveForce());
	
	Vector3D oldvelocity = new Vector3D(velocity); // v0 = v, make copy
	velocity.addInplace( // v += F * (dt/m) : from F = ma and v = at
	    forceAccumulator.multScalar(deltaTime/inertialMass));
	// x += (v + v0) / 2dt
	position.addInplace(velocity.add(oldvelocity).multScalar(deltaTime/2));
	
	forceAccumulator = new MutableVector3D(); // = 0
    }
	
	private void initDataMembers(BlockReader br) throws BRLoadingException
    // initialize data members from br, then initialize properties; no
    // need to override this in derived classes, instead, these classes
    // should call the inherited constructor if a new constructor is
    // defined at all
    {
        String[] brEntry; // array for entry read from br
        
        // let's get the name first, the name HAS TO BE PROVIDED, so we can go ahead and grab it!
        if ((br.hasMoreEntries()) && ((brEntry = br.getNextEntry())[0].equals("name")))
            {
                name = brEntry[1];
            }
        else // complain
            {
                throw new BREntryNotFoundException("name",
		                                   br.getEntryLocation());
            }
        
        // first set default values, 0, in case nothing gets set later
        position = new MutableVector3D();
        velocity = new MutableVector3D();
        inertialMass = 0;
        
        try {
            if (br.hasMoreEntries()) 
                {
                    brEntry = br.peekNextEntry();
                }
            else return; // if no more entries, terminate
            
            // try interpret next item as position
            if (brEntry[0].equals("position"))
                { 
                    position = new MutableVector3D(brEntry[1]);
                    br.getNextEntry(); // advance pointer
                    if (br.hasMoreEntries())
                        {
                            brEntry = br.peekNextEntry(); // what is next?
                        }
                    else return; // if no more entries, terminate
                }
				
            // try interpret next item as velocity
            if (brEntry[0].equals("velocity"))
                {
                    velocity = new MutableVector3D(brEntry[1]);
                    br.getNextEntry(); // advance pointer
                    if (br.hasMoreEntries())
                        {
                            brEntry = br.peekNextEntry(); // what is next?
                        }
                    else return; // if no more entries, terminate
                }

            // try interpret next item as inertialMass
            if (brEntry[0].equals("inertialMass"))
                {
                    br.getNextEntry(); // advance pointer
                    inertialMass = Double.parseDouble(brEntry[1]);
                }
        } // end try
        catch (ParseException e)
            {
                throw new BRLoadingParseException("Invalid Vector3D ", br.getEntryLocation());
            }
        catch (NumberFormatException e)
            {
                throw new BRLoadingParseException("Invalid Double", br.getEntryLocation());
            }
    }


    
    

    private void initproperties(BlockReader br) throws BRLoadingException
    {
	// list of all the laws held by physics
	Laws lawList = physics.getLaws();
	properties = new LawPropertyRecord[lawList.size()]; // create properties as an array of appropriate size

	BlockReader lawBlock;
	Law currentLaw;
	
	// first construct everything given in the BlockReader
	while (br.hasMoreBlocks())
	    {
		lawBlock = br.getNextBlock();
		try
		    {
			// get Law this subblock refers to
			currentLaw = lawList.getLaw(lawBlock.getHeader());

			properties[currentLaw.getNumber()] = currentLaw.constructNewPropertyRecord(lawBlock, this);
		    }
		catch (LawNotFoundException e)
		    {
			throw new BRLoadingException("Law " + lawBlock.getHeader() + " not found!", br.getBlockLocation());
		    }
	    }

	// now ask the corresponding Laws to provide defaults for everything we haven't constructed yet
	for (int i = 0; i < Array.getLength(properties); i++)
	    {
		if (properties[i] == null)
		    properties[i] = ((Law)lawList.get(i)).constructNewPropertyRecord(null, this);
	    }
    }
	
    /** Return this Entity's own contribution to the resultant force.
    This should be overridden for active objects (previously known as Intellegent objects).
    @returns In Entity, this method returns the zero vector Vector3D(0,0,0). */
    protected Vector3D getActiveForce()
    {
	return new Vector3D(); // 0
    }

    /** Returns a String representation of the Entity for debug-quality printing. */
    public String toString()
    {
	return ("name: " + getName() +
		"\nposition: " + getPosition().toString() +
		"\nvelocity: " + getVelocity().toString() +
		"\ninertialMass: " + getInertialMass() +
		"\nforceAccumulator: " + forceAccumulator);
    }
    
    /** This method has been added, because it is needed by certain laws like SphereCollision. */
    public void setVelocity(Vector3D newVelocity)
    {
	velocity = new MutableVector3D(newVelocity);
    }

    /** While I was at it, I've also added this method, in case it'll be needed later for some
	other law.*/
    public void setPosition(Vector3D newPosition)
    {
	position = new MutableVector3D(newPosition);
    }

    /** Returns the Physics object this Entity belongs to. */
    public Physics getPhysics()
    {
	return physics;
    }

    /** Disposes of system resources. Added by Miklos Reiter on 19-Apr-2001.*/
    public void dispose()
    {
	position = null;
	velocity = null;
	forceAccumulator = null;
	name = null;
	physics = null;
	// dispose of each of the LawPropertyRecord objects
	for (int i = 0; i < Array.getLength(properties); i++)
	    {
		properties[i].dispose();
	    }
	// now delete the array
	properties = null;
    }
}
