/**
 * Copyright (C) 2001  James Kermode (see the file README for details on
 * how to contact me)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package noope.input;

// for dynamically loading classes:
import java.lang.Class;
import java.lang.ClassNotFoundException;
import java.lang.LinkageError;
import java.lang.ExceptionInInitializerError;
import java.lang.reflect.Constructor;
import java.lang.NoSuchMethodException;
import java.lang.SecurityException;
import java.lang.NullPointerException;
import java.lang.reflect.InvocationTargetException;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;
import java.lang.IllegalArgumentException;


/** This class provides utility functions for dynamically loading classes
    from BlockReader objects. It is used by Laws, Entities and OutputProjection.
    All methods are static, you should not construct instances of this class.
    
    @author Originally written by Miklos Reiter as methods of OutputProjection,
            relocated by James Kermode.
    @version 12-Apr-2001 Changed by Miklos Reiter so the methods take BlockReaderLocation objects
    instead of BlockReaders, because this is all they need and to provide more flexibility.
 */
public class BRClassLoaderUtils
{
    /** Gets the class with the given name. br is needed to get the EntryLocation() if something
	fails. */
    public static Class getClassWithExactName(BlockReaderLocation brL, String className) throws
	BRLoadingException
    {
	try 
	    {
		return Class.forName(className);
	    }
	catch (ClassNotFoundException e) 
	    {
		// if this class was not found, throw an exception
		throw new BRLoadingException("Could not find the class " +
					     className + "!", 
					     brL);
	    }
	catch (ExceptionInInitializerError e) 
	    {
		throw new BRLoadingException("ExceptionInInitializerError occurred while " +
					     "trying to get Class object for " + className + 
					     ". The error message is: " + e + ". The exception"
					     + " occurred is: " + e.getException(),
					     brL);
	    } 
	catch (LinkageError e) 
	    {
		throw new BRLoadingException("LinkageError occurred while trying to get "
					     + "Class object for " + className + ". The error "
					     + "message is: " + e + ". Try recompiling your "
					     + "classes. For more information, read the Java "
					     + "API documentation.", brL);
	    }	
    }
    
    
    /** Tries to get the class with the given name. It tries the given name and the given name
	prefixed with defaultPackage. If something fails, an appropriate
	BRLoadingException occurrs. */
    public static Class getClassWithName(BlockReaderLocation brL, String className, 
					 String defaultPackage) throws
	BRLoadingException
    {
	try 
	    {
		try 
		    {
			return Class.forName(className);
		    }
		catch (ClassNotFoundException e) 
		    {
			// if the class was not found, try to find it in the package noope.output
			try 
			    {
				return Class.forName((defaultPackage + "." + className));
			    }
			catch (ClassNotFoundException f) 
			    { 
				// if this class was not found either, throw an exception
				throw new BRLoadingException("Could not find the class " +
							     className + "!", 
							     brL);
			    } 
		    } 
	    }
	catch (ExceptionInInitializerError e) 
	    {
		throw new BRLoadingException("ExceptionInInitializerError occurred while " +
					     "trying to get Class object for " + className + 
					     ". The error message is: " + e + ". The exception"
					     + " occurred is: " + e.getException(),
					     brL);
	    } 
	catch (LinkageError e) 
	    {
		throw new BRLoadingException("LinkageError occurred while trying to get "
					     + "Class object for " + className + ". The error "
					     + "message is: " + e + ". Try recompiling your "
					     + "classes. For more information, read the Java "
					     + "API documentation.", brL);
	    } 
    }


    /** This method asserts that the Class child is a subclass of the Class parent
	If it isn't, a BRLoadingException is thrown. */
    public static void assertSubClass(BlockReaderLocation brL, Class parent, Class child)
	throws BRLoadingException
    {
	// Now we can compare it with theClass:
	try 
	    {
		if (!parent.isAssignableFrom(child))
		    {
			throw new BRLoadingException("The class: " + child.getName() + " is " +
						     "not a subclass of " + parent.getName(),
						     brL);
		    }
	    }
	catch (NullPointerException e)
	    {
		// this should not happen!
		if (parent == null)
		    throw new BRLoadingException("Could not find class " + parent.getName() +
						 ", although doing so was apparently successful!",
						 brL);
		else // the child is null
		    throw new BRLoadingException("Could not find class " + child.getName() + 
						 ", although doing so was apparently successful!",
						 brL);
	    }
    }


    /** Gets the constructor of theClass matching the parameter format paramsClasses. */
    public static Constructor getConstructor(BlockReaderLocation brL, Class theClass, 
					     Class[] paramsClasses)
	throws BRLoadingException
    {
	try
	    {
		return theClass.getDeclaredConstructor(paramsClasses);
	    }
	catch (NoSuchMethodException e)
	    { // no such constructor exists
		throw new BRLoadingException("The class " + theClass.getName() + " has "
					     + "no constructor taking the requested parameters",
					     brL);
	    }
	catch (SecurityException e)
	    { // access to the information denied
		throw new BRLoadingException("SecurityException: access to the " +
					     "constructor denied!", brL);
	    }
    }

    /** Creates a new instance constructed with theConstructor, with params as the arguments. */
    public static Object getNewInstance(BlockReaderLocation brL, Class theClass, 
					Constructor theConstructor, Object[] params)
	throws BRLoadingException
    {
	try
	    {
		return theConstructor.newInstance(params);
	    }
	catch (InvocationTargetException e)
	    {
		Throwable f = e.getTargetException();
		
		if (f instanceof BRLoadingException)
		    {
			throw (BRLoadingException)f; // Throw on the BRLoadingException
		    }
		else // this is not something we'll be able to do something about, so we might
		    // as well print the stack trace.
		    {
			System.out.println("Exception in constructor: " + f);
			System.out.println("Printing stack trace...");
			f.printStackTrace();
			System.out.println("Throwing a BRLoadingException now...");
			throw new BRLoadingException("Exception in constructor: " + f,
						     brL);
		    }
	    }
	catch (ExceptionInInitializerError e)
	    {
		throw new BRLoadingException("ExceptionInInitializerError occurred while " +
					     "calling constructor for " + theClass.getName() +
					     ". The error message is: " + e + ". The " +
					     "exception occurred is: " + e.getException(),
					     brL);
	    }
	catch (InstantiationException e)
	    {
		throw new BRLoadingException("InstantiationException occurred while " +
					     "calling constructor. You are trying to " +
					     "instantiate an abstract class. Message: " +
					     e, brL);
	    }
	catch (IllegalAccessException e)
	    {
		throw new BRLoadingException("IllegalAccessException occurred while " +
					     "calling constructor. You may not have" +
					     " access to this class because it is private "
					     + "and in another package. Message: " +
					     e, brL);
	    }
	catch (IllegalArgumentException e)
	    {
		throw new BRLoadingException("Illegal number of arguments or wrong format"+
					     " or conversion failed. This should not " +
					     "happen! Message: " + e, brL);
	    }
    }
}
