/**
 * Copyright (C) 2001  Miklos Reiter (see the file README for details on
 * how to contact me)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package noope.laws; 

import noope.core.*;
import noope.input.BlockReader;
import noope.input.BRLoadingException;
import noope.entities.Entity;

/** This class describes a force-law in the modelled universe.
  Below is a skeleten example of how Law could be subclassed to create a useful law, for a full
  implementation see the {@link noope.laws.Gravity} class.
     <pre>

class GravityPropertyRecord extends LawPropertyRecord
{
  GravityPropertyRecord(BlockReader, Entity) // get gravitation mass from BlockReader
  protected double gravitationalMass;
  public double getGravitationalMass();
}

class Gravity extends Law
{
  Gravity(BlockReader br, int number, Physics phys)
    {
       super(br, number);
       // do rest of initialization from br
    }

  LawPropertyRecord constructNewPropertyRecord(BlockReader r, Entity e)
    { 
       return new GravityPropertyRecord(r, e);
    }
  
  public void act()
    { // for each entity in Entities
      // work out force on it and call entity.addForce, by using entity.getPropertyRecord(lawnum)
    }
  ...
}
     </pre>
	 
	 @author Miklos Reiter
	 @version 19-Apr-2001 Added dispose() [Miklos Reiter]
 */
public abstract class Law 
{
    private String name; // stores the name of the Law, as used for initialization
    private int number; // stores the number of the Law, which is its index in Laws and also the index
    // of the corresponding LawPropertyRecord in Entity.properties.
   
    /** Constructs a new Law object from the supplied BlockReader, with lawnumber as its number.
     * These numbers are allocated by the Laws constructor. Everything apart from the number is loaded from br. 
     * The constructor of every derived law should follow this interface.
     * Subclasses should call the inherited constructor first to make sure the name is assigned. 
     * Also, subclasses, should have a constructor taking a BlockReader and an int, so it can be called by Physics, 
     * even though it does not know about the class.
     * 
     * The entries in br have to be as follows:
      <pre>
      [name "lawName"]
          (if name not given as first entry, will assume classname, i.e. parameter as the name of the law)
      ... (custom entries for derived laws)
      ...
      </pre>
     * @param br BlockReader used to initialise object
     * @param lawnumber Number of this law
     * @see noope.core.Laws#Laws(BlockReader)
     */

    /** A reference to the physics world so getActiveForce can learn about its
    environment. */
    protected Physics physics;
    
    /** Constructs a new Law object from the supplied BlockReader, with lawnumber as its number.
     * These numbers are allocated by the Laws constructor. Everything apart from the number is loaded from br. 
     * The constructor of every derived law should follow this interface.
     * Subclasses should call the inherited constructor first to make sure the name is assigned. 
     * Also, subclasses, should have a constructor taking a BlockReader and an int, so it can be called by Physics, 
     * even though it does not know about the class.
     * 
     * The entries in br have to be as follows:
      <pre>
      [name "lawName"]
          (if name not given as first entry, will assume classname, i.e. parameter as the name of the law)
      ... (custom entries for derived laws)
      ...
      </pre>
     * @param br BlockReader used to initialise object
     * @param lawnumber Number of this law
     * @param phys Reference to the Physics world
     * @see noope.core.Laws#Laws(BlockReader, Physics)
     */
    public Law(BlockReader br, int lawnumber, Physics phys) throws BRLoadingException
    {
	String[] brEntry;
	number = lawnumber;
	physics = phys;

	// if the law has entries and the first entry has the title "name" then use its argument
	// as the name of the law
	// first peek only! changed by Miklos Reiter on 28 March 2001
	if (br.hasMoreEntries() && ((brEntry = br.peekNextEntry())[0].equals("name"))) // short-circuit evaluation
	    {
		name = brEntry[1];
		br.getNextEntry(); // now move on to next entry
	    }
	else // take the parameter
	    {
		name = br.getParameter();
	    }
    }
    
    /** Returns the name of the law object. This is as specified in the BlockReader object from which this Law object has
    been constructed. */
    public String getName()
    {
	return name;
    }

    /** Returns the number of the law object. The number is the index of this Law in Laws and also the index of the
    corresponding LawPropertyRecord in Entity.properties */
    public int getNumber()
    {
	return number;
    }

    /** Make the law act upon the entities. This is called by Physics once in every time interval.
     * This method should calculate the net force due to this Law acting on each Entity and then call
     *    Entity.addForce(force)
     * on every Entity to add this law's constribution to the total force acting on the Entity.
     * @param e the list of entities that the force should act on
     * @see noope.entities.Entity#addForce(Vector3D)
     */
    abstract public void act();

    /** Create a new LawPropertyRecord from the BlockReader object. This is done
      by calling the corresponding LawPropertyRecord's constructor with the BlockReader object. 
      Most subclasses will want to override this method to construct an instance of their own 
      LawPropertyRecord class, but this is not a requirement. If a law does not need to store 
      any data, it can use the default method that just returns an instance of LawPropertyRecord. 
      If this is what you want, pass it a null BlockReader object, or it will issue a warning.
      
      @param br BlockReader used to read properties from
      @param ent is the instance of Entity, this LawPropertyRecord will be constructed for
     */
    public LawPropertyRecord constructNewPropertyRecord(BlockReader br, Entity ent) throws
	BRLoadingException
    {
	if (br != null)
	    {
		// Tell the user that they are constructing an empty LawPropertyRecord from a non-null
		// BlockReader. Is this a mistake?
		System.out.println("TODO: Warning: Law.constructNewPropertyRecord: Are you sure you want to create "
				   + "an empty LawPropertyRecord? You have provided a non-null BlockReader!");
	    }
	return new LawPropertyRecord(br, ent);
    }

    /** Provides for debug-quality printing. Print the name and number of this Law.*/
    public String toString()
    {
	return "name: " + getName() + "\nnumber: " + getNumber();
    }

    /** Dispose of system resources used by this law. This can also be used for output laws to
	close any output devices like files, windows, etc. used. Added by Miklos Reiter on
	19-Apr-2001. */
    public void dispose()
    {
	name = null;
	physics = null;
    }
}
