/**
 * Copyright (C) 2001  Miklos Reiter (see the file README for details on
 * how to contact me)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package noope.output;

// some noope-related classes
import noope.core.Physics;
import noope.core.Entities;
import noope.core.Vector3D;

import noope.entities.Entity;

import noope.laws.Law;
import noope.laws.LawPropertyRecord;

import noope.input.BlockReader;
import noope.input.BlockReaderLocation;
import noope.input.BRLoadingException;
import noope.input.BRLoadingParseException;
import noope.input.BREntryNotFoundException;
import noope.input.BRClassLoaderUtils;

// some standard library classes
import java.lang.Integer;
import java.lang.Double;
import java.lang.NumberFormatException;
import java.text.ParseException;
import java.util.ListIterator;

// For dynamically loading classes
import java.lang.reflect.Constructor;

// some drawing classes
import java.awt.image.BufferedImage;
import java.awt.Graphics2D; // a Graphics object for drawing.
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.geom.AffineTransform;


/** This is an output class that can be used to display in real-time an orthogonal projection onto
    a plane of all the Entity objects in Entities.

    The plane is specified by its normal (the projection onto all such planes is the same) and is
    then chosen to pass through the origin for simplicity. If the plane is the x-y plane, the x and
    y axes are used as the coordinate axes. Otherwise, the x-axis is the intersection of the plane
    with the x-y plane, the origin is the origin of the 3-d space we are representing and the
    y-axis is perpendicular to the x-axis. The positive x direction is the one which has a
    positive x component in the old coordinates. The positive y direction is the vector/cross
    product of the unit-normal and the new x unit vector.

    @author Miklos Reiter, this class is based on OutputText by James Kermode
    @version 26-Apr-2001 changed to use BufferedFrame [Miklos Reiter], a subclass of Frame, so
    we can use buffering with AWT; and added a call to windowFrame.pack() before resizing it, so
    it can determine the width needed for the border
    22-Apr-2001: changed to access drawingWindow in a synchronized way [Miklos Reiter]
*/
public class OutputProjection extends Law
{
    // Protected variables:

    /** The name of the gif file used for outputting the animation, null if animation is
	displayed directly on screen. */
    protected String gifName;

    /** String representing the comments written into the gif file. Only meaningful if gifName is
	set. */
    protected String gifComments = "";

    /** How many times to play sequence. Written into gif. Default is 1, 0 represents indefinite
	looping. */
    protected int gifLoopCount = 1;

    /** Number of centiseconds to wait between two frames. Default = 0. Written into gif. */
    protected int gifUniformDelay = 0;

    /** True after the window has been disabled (by the user closing it or by dispose()). */
    protected boolean disabled = false;

    /** The time between two consecutive outputs. */
    protected double period;

    /** Last time an output was made. Output will occur the first time the act method is called.*/
    protected double lastTime = Double.NaN;
    
    /** The AnimationListener used to display the animation. */
    protected AnimationListener outputObject;
    
    /** The JFrame used to display window. Only used if gifName == null. */
    protected BufferedFrame windowFrame;


    /** The width of the image we are displaying. */
    protected int width;
    /** The height of the image we are displaying. */
    protected int height;
    /** The x-coordinate of the position at which we are displaying the window. */
    protected int xpos;
    /** The y-coordinate of the position at which we are displaying the window. */
    protected int ypos;
    
    /** xmin, xmax, ymin and ymax describe the range of new coordinates to draw. */
    protected double xmin, xmax, ymin, ymax;
    
    /** The unit normal to the plane we are projecting on. */
    protected Vector3D normal;
    
    /** The new x and y unit vectors. These give us the direction of the positive coordinate
	axes in the plane we are projecting on. xnew and ynew are orthogonal and are oriented
	according to the right-handed corkscrew rule with respect to normal.
     */
    protected Vector3D xnew, ynew;


    // Public static variables: provide default values.

    /** The height of the image used by default if nothing is provided in the BlockReader object
	passed to the constructor. */
    public static int defaultWidth = 300;
    /** The width of the image used by default if nothing is provided in the BlockReader object
	passed to the constructor. */
    public static int defaultHeight = 300;

    /** The default x position at which the window should appear. Can be overridden by giving
	a value explicitly in the BlockReader object passed to the constructor.*/
    public static int defaultXpos = 0;
    /** The default y position at which the window should appear. Can be overridden by giving
	a value explicitly in the BlockReader object passed to the constructor.*/
    public static int defaultYpos = 0;


    /** Constructs a new OutputProjection object from the given BlockReader with the given
	lawnumber. universe is stored as the reference to the Physics object hosting this
	simulation.
	br Should contain a double "period", specifying the period of sampling, followed by 
	"normal", the normal unit vector to the plane onto which we wish to project, doubles "xmin",
	"xmax", "ymin" and "ymax" specifying the range in the new coordinates to be represented;
        integers "width" and "height" specifying the width and the height in pixels of the 
	image to use. "width" and "height" may be omitted, in which case the default is
	defaultWidth and defaultHeight.
	
	Next, optionally "gifName", a filename specifying the file write the animated gif to.
	If gifName is not provided, the animation is going to be displayed in a window.
	If gifName is given, the following options can be specified:
	"gifComments": comments written into gif
	"gifUniformDelay": time in centiseconds to wait between subsequent frames (default = 0)
	"gifLoopCount": number of times animation should be played (default = 1, 0 = indefinite)
	
	Finally, integers "xpos", "ypos" specify the position at which the window should appear
	(optional, take defaultXpos, defaultYpos if not given). (only if gifName is not provided)
    */
    public OutputProjection(BlockReader br, int lawnumber, Physics phys) 
	throws BRLoadingException
    {
	super(br, lawnumber, phys);
	loadDataMembers(br);
	setupGraphics();
	calculateParameters();
	// we may now draw into the window
	disabled = false;
    }

    /**
     * Clears the background and sets the coordinate transform of g, so it can be used to display
     * the animation.
     */
    private void adjustGraphics(Graphics2D g)
    {
	// clear window before setting new coordinates
	g.setBackground(Color.white);
	g.clearRect(0, 0, width, height);
	
	// now set new coordinates for Graphics object, corresponds to the following:
	/*int outX = (int)(width*(x - xmin)/(xmax - xmin));
	  int outY = height - (int)(height*(y - ymin)/(ymax - ymin));*/
	g.setTransform(new 
	    AffineTransform(((double)width)/((double)(xmax - xmin)), 0.0, 0.0, 
			    -((double)height)/((double)(ymax - ymin)),
			    -((double)(width*xmin))/((double)(xmax - xmin)),
			    ((double)(height*ymax))/((double)(ymax - ymin))));	
    }

    private void setupGraphics()
    {
	// type of image
	int imageType = BufferedImage.TYPE_3BYTE_BGR; // type of image

	BufferedImage im = new BufferedImage(width, height, imageType);
	
	if (gifName == null) // create objects to display animation
	    {
		outputObject = new DrawingWindow(im);
		adjustGraphics(outputObject.drawingGraphics());
		windowFrame = new BufferedFrame("Projection: " + getName(), width, height);

		// dispose window on close and hide it
		windowFrame.addWindowListener(new WindowAdapter()
		    {
			public void windowClosing(WindowEvent e)
			{
			    synchronized (outputObject)
				{
				    outputObject.dispose();
				    windowFrame.setVisible(false);
				}
			}
		    });
		
		windowFrame.add((DrawingWindow)outputObject, BorderLayout.CENTER);
		
		windowFrame.pack();
		
		windowFrame.setLocation(xpos, ypos);
		windowFrame.setVisible(true);
	    }
	else // create objects to encode gif
	    {
		outputObject = new AnimatedGifCreator(im, gifName, gifComments,
						      gifUniformDelay, gifLoopCount);
		adjustGraphics(outputObject.drawingGraphics());
	    }
	
    }

    // Do some initial calculations:
    private void calculateParameters()
    {
	if (normal.getX() != 0.0) // x-component non-zero
	    {
		xnew = (new Vector3D(normal.getY()/normal.getX(), -1.0, 0.0)).getUnitVector();
		if (xnew.getX() < 0) xnew = xnew.negative();
		ynew = normal.crossProduct(xnew);
	    }
	else
	    {
		if (normal.getY() != 0.0) // y-component non-zero
		    {
			xnew = (new Vector3D(1.0, -normal.getX()/normal.getY(),0.0))
			    .getUnitVector();
			ynew = normal.crossProduct(xnew);
		    }
		else // normal parallel to z-axis
		    {
			if (normal.getZ() > 0.0)
			    {
				xnew = Vector3D.ex;
				ynew = Vector3D.ey;
			    }
			else if (normal.getZ() < 0.0)
			    {
				xnew = Vector3D.ex;
				ynew = Vector3D.ey.negative();
			    }
		    }
	    }
    }

    private void loadDataMembers(BlockReader br) throws BRLoadingException
    {
	String[] brEntry = null;

	// Get period, we don't have to use the careful peek..., because if the period is
	// missing, we are bailing out, anyway.
	if (br.hasMoreEntries() && ((brEntry = br.getNextEntry())[0].equals("period")))
        {
            try
	    {
		period = Double.valueOf(brEntry[1]).doubleValue();
            }
            catch (NumberFormatException e)
            {
		throw (new
		    BRLoadingParseException("Failed to parse for double" , br.getEntryLocation()));
            }
	}
        else
        {
	    System.out.println(br.hasMoreEntries() + brEntry[0] + brEntry[1]);
	    throw (new
		BREntryNotFoundException("period", br.getEntryLocation()));
        }


	// now get normal, Vector3D
	if (br.hasMoreEntries() && ((brEntry = br.getNextEntry())[0].equals("normal")))
        {
            try
	    {
		normal = new Vector3D(brEntry[1]);
		if (normal.getX() == 0.0 && normal.getY() == 0.0 && normal.getZ() == 0.0)
		    {
			throw new BRLoadingException("The given vector is zero, give a non-zero "
						     + "vector for the normal direction.",
						     br.getEntryLocation());
		    }
		else
		    {
			normal = normal.getUnitVector();
		    }
            }
            catch (ParseException e)
            {
		throw (new
		    BRLoadingParseException("Failed to parse for Vector3D" , 
					    br.getEntryLocation()));
            }
	}
        else
        {
	    throw (new
		BREntryNotFoundException("normal", br.getEntryLocation()));
        }


	// get xmin
	if (br.hasMoreEntries() && ((brEntry = br.getNextEntry())[0].equals("xmin")))
        {
            try
	    {
		xmin = Double.valueOf(brEntry[1]).doubleValue();
            }
            catch (NumberFormatException e)
            {
		throw (new
		    BRLoadingParseException("Failed to parse for double" , br.getEntryLocation()));
            }
	}
        else
        {
	    throw (new
		BREntryNotFoundException("xmin", br.getEntryLocation()));
        }


	// get xmax
	if (br.hasMoreEntries() && ((brEntry = br.getNextEntry())[0].equals("xmax")))
        {
            try
	    {
		xmax = Double.valueOf(brEntry[1]).doubleValue();
            }
            catch (NumberFormatException e)
            {
		throw (new
		    BRLoadingParseException("Failed to parse for double" , br.getEntryLocation()));
            }
	}
        else
        {
	    throw (new
		BREntryNotFoundException("xmax", br.getEntryLocation()));
        }

	// get ymin
	if (br.hasMoreEntries() && ((brEntry = br.getNextEntry())[0].equals("ymin")))
        {
            try
	    {
		ymin = Double.valueOf(brEntry[1]).doubleValue();
            }
            catch (NumberFormatException e)
            {
		throw (new
		    BRLoadingParseException("Failed to parse for double" , br.getEntryLocation()));
            }
	}
        else
        {
	    throw (new
		BREntryNotFoundException("ymin", br.getEntryLocation()));
        }

	// get ymax
	if (br.hasMoreEntries() && ((brEntry = br.getNextEntry())[0].equals("ymax")))
        {
            try
	    {
		ymax = Double.valueOf(brEntry[1]).doubleValue();
            }
            catch (NumberFormatException e)
            {
		throw (new
		    BRLoadingParseException("Failed to parse for double" , br.getEntryLocation()));
            }
	}
        else
        {
	    throw (new
		BREntryNotFoundException("ymax", br.getEntryLocation()));
        }

	// get width (optional, so use peek...)
	if (br.hasMoreEntries() && ((brEntry = br.peekNextEntry())[0].equals("width")))
	    {
		try
		    {
			br.getNextEntry(); // advance pointer
			width = Integer.parseInt(brEntry[1]);
		    }
		catch (NumberFormatException e)
		    {
			throw (new
			    BRLoadingParseException("Failed to parse for int" , 
						    br.getEntryLocation()));
		    }
	    }
	else // don't complain if nothing is given, use default!
	    {
		width = defaultWidth;
	    }


	// get height (optional, so use peek...)
	if (br.hasMoreEntries() && ((brEntry = br.peekNextEntry())[0].equals("height")))
	    {
		try
		    {
			br.getNextEntry(); // advance pointer
			height = Integer.parseInt(brEntry[1]);
		    }
		catch (NumberFormatException e)
		    {
			throw (new
			    BRLoadingParseException("Failed to parse for int" , 
						    br.getEntryLocation()));
		    }
	    }
	else // don't complain if nothing is given, use default!
	    {
		height = defaultHeight;
	    }
	
	// do we have to output to a gif? if yes, to what fileName?
	if (br.hasMoreEntries() && ((brEntry = br.peekNextEntry())[0].equals("gifName")))
	    {
		br.getNextEntry(); // advance pointer
		gifName = brEntry[1];

		// ok, so we've found gifName, look for other gif-related stuff...

		// gifComments
		if (br.hasMoreEntries() && ((brEntry = br.peekNextEntry())[0].equals("gifComments")))
		    {
			br.getNextEntry(); // advance pointer
			gifComments = brEntry[1];
		    }

		// gifUniformDelay
		if (br.hasMoreEntries() && ((brEntry = br.peekNextEntry())[0].equals("gifUniformDelay")))
		    {
			try
			    {
				br.getNextEntry(); // advance pointer
				gifUniformDelay = Integer.parseInt(brEntry[1]);
			    }
			catch (NumberFormatException e)
			    {
				throw (new
				    BRLoadingParseException("Failed to parse for int" ,
							    br.getEntryLocation()));
			    }
		    }

		// gifLoopCount
		if (br.hasMoreEntries() && ((brEntry = br.peekNextEntry())[0].equals("gifLoopCount")))
		    {
			try
			    {
				br.getNextEntry(); // advance pointer
				gifLoopCount = Integer.parseInt(brEntry[1]);
			    }
			catch (NumberFormatException e)
			    {
				throw (new
				    BRLoadingParseException("Failed to parse for int" ,
							    br.getEntryLocation()));
			    }
		    }
	    }
	else
	    {
		// we'll have to display everything in a window, so have a look, whether arguments
		// are provided for that
		
		// get xpos (optional, so use peek...)
		if (br.hasMoreEntries() && ((brEntry = br.peekNextEntry())[0].equals("xpos")))
		    {
			try
			    {
				br.getNextEntry(); // advance pointer
				xpos = Integer.parseInt(brEntry[1]);
			    }
			catch (NumberFormatException e)
			    {
				throw (new
				    BRLoadingParseException("Failed to parse for int" , 
							    br.getEntryLocation()));
			    }
		    }
		else // don't complain if nothing is given, use default!
		    {
			xpos = defaultXpos;
		    }

		// get ypos (optional, so use peek...)
		if (br.hasMoreEntries() && ((brEntry = br.peekNextEntry())[0].equals("ypos")))
		    {
			try
			    {
				br.getNextEntry(); // advance pointer
				ypos = Integer.parseInt(brEntry[1]);
			    }
			catch (NumberFormatException e)
			    {
				throw (new
				    BRLoadingParseException("Failed to parse for int" , 
							    br.getEntryLocation()));
			    }
		    }
		else // don't complain if nothing is given, use default!
		    {
			ypos = defaultYpos;
		    }
	    }
    }

    /** Does not exert a force on any objects, just outputs. */
    public void act()
    {
	// Do we need to output now? this will be true the first time, since a comparison with a
	// NaN is always false!
	if ((!disabled) && !(physics.getTime() - lastTime < period))
	    {
		synchronized (outputObject) {
		    
		    if (!outputObject.isWorking())
			{
			    dispose();
			    return;
			}
		    
		    // clear background
		    outputObject.drawingGraphics().setColor
			(outputObject.drawingGraphics().getBackground());
		    outputObject.drawingGraphics().fill(new java.awt.geom.Rectangle2D.Double
			(xmin, ymin, xmax - xmin, ymax - ymin));
		    
		    
		    Entities entities = physics.getEntities();
		    Entity entity;
		    Vector3D pos;
		    
		    // For each entity in Entities, output position
		    for (ListIterator i = entities.listIterator(); i.hasNext(); )
			{
			    entity = (Entity)i.next();
			    pos = entity.getPosition();
			    
			    // calculate position on Projection:
			    // projection = x - (x.n)n
			    Vector3D projection = 
				pos.subtract(normal.multScalar(pos.dotProduct(normal)));
			    // x = projection . xnew; y = projection . ynew
			    double x = projection.dotProduct(xnew);
			    double y = projection.dotProduct(ynew);
			    
			    // draw position into drawingGraphics
			    outputPosition(entity, x, y);
			}
		    lastTime = physics.getTime(); // update lastTime
		    
		    // animation frame complete!
		    outputObject.updateComplete();
		    
		} // end synchronized
	    }
    }


    /** Outputs the position of the Entity e at x, y in the coordinates of the plane we are
	projecting onto. Just gets the Entity's OutputProjectionPropertyRecord and calls its
	paint method with drawingGraphics and the coordinates. */
    protected void outputPosition(Entity e, double x, double y)
    {
	// ask the object to draw itself
	((OutputProjectionPropertyRecord)e.getPropertyRecord(getNumber()))
	    .paint(outputObject.drawingGraphics(), x, y);
    }



    /** Constructs a new OutputProjectionPropertyRecord object. If the parameter of br is null
	or an empty String, then an object of the class OutputProjectionPropertyRecord is
	constructed. If, however, a parameter is given, we try to get the subclass of
	OuputProjectionPropertyRecord with that name. First the name itself is tried and if
	no class by that name is found, we try to find it in the package noope.laws. */
    public LawPropertyRecord constructNewPropertyRecord(BlockReader br, Entity ent)
                                             throws BRLoadingException
    {
	LawPropertyRecord result;
	String className = null;
	BlockReaderLocation brL = null; // for errors

	if (br != null) className = br.getParameter();
	if (className == null || className == "")
	    {
		result = new OutputProjectionPointPropertyRecord(br, ent);
	    }
	else
	    {
		brL = br.getEntryLocation();
		result = ent.getPhysics().getLawPropertyRecordLoader().createLawPropertyRecord
		    (className, brL, br, ent);
	    }
	
	if (br != null) br.assertEmpty(); // make sure, nothing is left in the BlockReader object
	
	try
	    {
		return (OutputProjectionPropertyRecord)result;
	    }
	catch (java.lang.ClassCastException e)
	    {
		throw new BRLoadingException("Class " + className + " is not a subclass of "
					     + "OutputProjectionPropertyRecord.", brL);
	    }
    }

    /** Closes the windows opened, then calls the inherited method. */
    public void dispose()
    {
	// don't draw into it any more
	disabled = true;

	if (windowFrame != null)
	    {
		// hide the window
		windowFrame.setVisible(false);
		// now dispose it
		windowFrame.dispose();
	    }
	outputObject.dispose();
	// call inherited method
	super.dispose();
    }
}
