/**
 * Copyright (C) 2001  James Kermode (see the file README for details on
 * how to contact me)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package test;

/* TestOutputText - James Kermode, 26-Mar-2001
 * further modified to test the OutputText Law by James Kermode 22-Mar-2001
 * modified from TestPhysics by James Kermode
 *
 * Tests the physics engine using a simulation of the Earth's motion around the Sun.
 *
 * DATA:
 * Sun: at origin, initial velocity = 0, Mass = 100
 * Earth: at (1, 0, 0), initial velocity (0, 10, 0), Mass = 1
 *
 * Gravity: G = 1
 *
 * Observe orbit is nearly circular. It becomes nearer to circular, as we increase the Sun's inertialMass, without
 * changing its gravitationalMass.
 *
 */

import java.util.ListIterator;

import noope.core.*;
import noope.input.*;

public class TestOutputText
{
    public static void main(String[] argv) throws BRLoadingException
    {
	final double deltat = 0.0001; // timestep
	
	// Set up an ArrayBlockReader to load physics from
	BlockReader[] lawsBlocks = {newGravBr(), newOutputTextBr()};
	BlockReader[] entitiesBlocks = {newEntity1Br(), newEntity2Br()};
	BlockReader lawsBr = newLawsBr(lawsBlocks);
	BlockReader entitiesBr = newEntitiesBr(entitiesBlocks);
	BlockReader physicsBr = newPhysicsBr(lawsBr, entitiesBr);
	
	// Construct physics from this ArrayBlockReader
	Physics physics = new Physics(physicsBr);
	
	System.out.println("Running simulation for 1 second...");
	for (int i = 0; i < 10000; i++)
  	   physics.step(deltat);
    }
    
    // *** Miklos' ArrayBlockReader test functions:
    
    private static BlockReader newGravBr() // creates an ArrayBlockReader defining a Gravity law
    {
	String header = "law", parameter = "Gravity"; 
	// these aren't really needed, but every BlockReader has to have
	// them
	String[][] entries = 
	{{"name", "Gravity"},
	 {"G", "1"}
	};
	BlockReader[] subblocks = {};
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }

    private static BlockReader newOutputTextBr() // creates an ArrayBlockReader defining an OutputTextlaw
    {
	String header = "law", parameter = "noope.output.OutputText"; 
	// these aren't really needed, but every BlockReader has to have
	// them
	String[][] entries = 
	{{"name", "OutputText"},
	 {"file", "sun_earth.txt"},
	 {"interval", "100"}
	};
	BlockReader[] subblocks = {};
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }    
    
    private static BlockReader newEntity1Br() // creates an ArrayBlockReader defining an Entity
    {
	String header = "entity", parameter = ""; // these aren't really needed, but every BlockReader has to have
	// them
	String[][] entries = 
	{{"name", "Earth"},
	 {"position", "(1, 0, 0)"},
	 {"velocity", "(0, 10, 0)"},
	 {"inertialMass", "1"}
	};
	BlockReader[] subblocks = {}; // don't give it a GravityPropertyRecord
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }
    
    private static BlockReader newEntity2Br() // creates an ArrayBlockReader defining an Entity
    {
	String header = "entity", parameter = ""; // these aren't really needed, but every BlockReader has to have
	// them
	String[][] entries = 
	{{"name", "Sun"},
	 {"position", "(0, 0, 0)"},
	 {"velocity", "0"},
	 {"inertialMass", "100"}
	};
	BlockReader[] subblocks = {newGravPropBr()}; // give it a GravityPropertyRecord
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }
    
    private static BlockReader newGravPropBr() // creates an ArrayBlockReader defining a GravityPropertyRecord
    {
	String header = "Gravity", parameter = null; 
	String[][] entries = {
	    {"gravitationalMass", "100"} // use same gravitationalMass as inertialMass, if you don't want to confuse
	    // people:-)
	};
	BlockReader[] subblocks = {};
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }
    
    // *** ArrayBlockReader functions added by James Kermode
    
    // creates the entities {...} block containing all the entities as subblocks
    private static BlockReader newEntitiesBr(BlockReader[] subblocks)
    {
	String header = "entities", parameter = null;
	String[][] entries = null;
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }    
    
    // creates the laws {...} block containing all the laws as subblocks
    private static BlockReader newLawsBr(BlockReader[] subblocks)
    {
	String header = "laws", parameter = null;
	String[][] entries = null;
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }    
    
    // creates the physics {...} block containing laws {...} and entities{...} as subblocks
    private static BlockReader newPhysicsBr(BlockReader laws, BlockReader entities)
    {
	String header = "physics", parameter = null;
	String[][] entries = null;
	BlockReader[] subblocks = {laws, entities};
	
	return new ArrayBlockReader(header, parameter, entries, subblocks);
    }    
}
